from dodgy.run import run_checks
import os
import re
from prospector.message import Location, Message
from prospector.tools.base import ToolBase


def module_from_path(path):
    # note : assumes a relative path
    module = re.sub(r'\.py', '', path)
    return '.'.join(module.split(os.path.sep)[1:])


class DodgyTool(ToolBase):

    def prepare(self, rootpath, ignore, args, adaptors):
        self.rootpath = rootpath

    def run(self):
        warnings = run_checks(self.rootpath)
        messages = []

        for warning in warnings:
            path = warning['path']
            loc = Location(path, module_from_path(path), '', warning['line'], 0, absolute_path=False)
            msg = Message('dodgy', warning['code'], loc, warning['message'])
            messages.append(msg)

        return messages



import sys

from collections import defaultdict

from .text import _SUMMARY_TEMPLATE


__all__ = (
    'format_messages',
)


def format_messages(summary, messages):
    if summary:
        sys.stdout.write("Check Information\n=================\n")
        sys.stdout.write(_SUMMARY_TEMPLATE % summary)
        sys.stdout.write('\n\n')

    if not messages:
        return

    # pylint: disable=W0108
    groups = defaultdict(lambda: defaultdict(list))

    for message in messages:
        groups[message.location.path][message.location.line].append(message)

    sys.stdout.write("Messages\n========\n\n")
    for filename in sorted(groups.keys()):
        sys.stdout.write('%s\n' % filename)

        for line in sorted(groups[filename].keys(), key=lambda x: 0 if x is None else int(x)):
            sys.stdout.write('  Line: %s\n' % line)

            for msg in groups[filename][line]:
                sys.stdout.write(
                    '    %(source)s: %(code)s / %(message)s' % msg.as_dict(),
                )
                if msg.location.character:
                    sys.stdout.write(' (col %s)' % msg.location.character)
                sys.stdout.write('\n')

        sys.stdout.write('\n')




from pep8 import StyleGuide, BaseReport, register_check
from pep8ext_naming import NamingChecker

from prospector.message import Location, Message
from prospector.tools.base import ToolBase


__all__ = (
    'Pep8Tool',
)


class ProspectorReport(BaseReport):
    def __init__(self, *args, **kwargs):
        super(ProspectorReport, self).__init__(*args, **kwargs)
        self._prospector_messages = []

    def error(self, line_number, offset, text, check):
        code = super(ProspectorReport, self).error(
            line_number,
            offset,
            text,
            check,
        )
        if code is None:
            # The error pep8 found is being ignored, let's move on.
            return
        else:
            # Get a clean copy of the message text without the code embedded.
            text = text[5:]

        # mixed indentation (E101) is a file global message
        if code == 'E101':
            line_number = None

        # Record the message using prospector's data structures.
        location = Location(
            path=self.filename,
            module=None,
            function=None,
            line=line_number,
            character=(offset + 1),
        )
        message = Message(
            source='pep8',
            code=code,
            location=location,
            message=text,
        )

        self._prospector_messages.append(message)

    def get_messages(self):
        return self._prospector_messages


class ProspectorStyleGuide(StyleGuide):
    def __init__(self, *args, **kwargs):
        # Remember the ignore patterns for later.
        self._ignore_patterns = kwargs.pop('ignore_patterns', [])

        # Override the default reporter with our custom one.
        kwargs['reporter'] = ProspectorReport

        super(ProspectorStyleGuide, self).__init__(*args, **kwargs)

    def excluded(self, filename, parent=None):
        if super(ProspectorStyleGuide, self).excluded(filename, parent):
            return True

        # If the file survived pep8's exclusion rules, check it against
        # prospector's patterns.
        if any([ip.search(filename) for ip in self._ignore_patterns]):
            return True

        return False


class Pep8Tool(ToolBase):
    def __init__(self, *args, **kwargs):
        super(Pep8Tool, self).__init__(*args, **kwargs)
        self.checker = None

    def prepare(self, rootpath, ignore, args, adaptors):
        # Instantiate our custom pep8 checker.
        self.checker = ProspectorStyleGuide(
            paths=[rootpath],
            ignore_patterns=ignore,
        )

        # Make sure pep8's code ignores are fully reset to zero.
        # pylint: disable=W0201
        self.checker.select = ()
        self.checker.ignore = ()

        # Let the adaptors & profiles do their thing.
        for adaptor in adaptors:
            adaptor.adapt_pep8(self.checker)

    def run(self):
        report = self.checker.check_files()
        return report.get_messages()


# Load pep8ext_naming into pep8's configuration.
register_check(NamingChecker)



import sys
import argparse
import os
import re
from datetime import datetime
from prospector.adaptor import LIBRARY_ADAPTORS
from prospector.adaptor.common import CommonAdaptor
from prospector.adaptor.profile import ProfileAdaptor
from prospector.autodetect import autodetect_libraries
from prospector.formatters import FORMATTERS
from prospector import tools, blender
from prospector import __pkginfo__
from prospector.message import Location, Message


def make_arg_parser():
    version = __pkginfo__.get_version()
    parser = argparse.ArgumentParser(
        description="Version %s. Performs analysis of Python code" % version
    )

    parser.add_argument(
        '-A', '--no-autodetect', action='store_true', default=False,
        help='Turn off auto-detection of frameworks and libraries used. By'
        ' default, autodetection will be used. To specify manually, see'
        ' the --uses option.',
    )

    parser.add_argument(
        '-B', '--no-blending', action='store_true', default=False,
        help="Turn off blending of messages. Prospector will merge together"
        " messages from different tools if they represent the same error."
        " Use this option to see all unmerged messages."
    )

    parser.add_argument(
        '-D', '--doc-warnings', action='store_true', default=False,
        help="Include warnings about documentation.",
    )

    parser.add_argument(
        '-M', '--messages-only', default=False, action='store_true',
        help="Only output message information (don't output summary"
        " information about the checks)",
    )

    output_help = "The output format. Valid values are %s" % (
        ', '.join(sorted(FORMATTERS.keys())),
    )
    parser.add_argument(
        '-o', '--output-format', default='text', help=output_help
    )

    parser.add_argument(
        '-p', '--path',
        help="The path to the python project to inspect (defaults to PWD)",
    )

    profiles_help = "The list of profiles to load. A profile is a certain" \
        " 'type' of behaviour for prospector, and is represented by a YAML" \
        " configuration file. A full path to the YAML file describing the" \
        " profile must be provided. (see --strictness)"
    parser.add_argument(
        '-P', '--profiles', default=[], nargs='+', help=profiles_help,
    )

    strictness_help = 'How strict the checker should be. This affects how' \
        ' harshly the checker will enforce coding guidelines. The default' \
        ' value is "medium", possible values are "veryhigh", "high",' \
        ' "medium", "low" and "verylow".'
    parser.add_argument(
        '-s', '--strictness', help=strictness_help, default='medium',
    )

    parser.add_argument(
        '-S', '--summary-only', default=False, action='store_true',
        help="Only output summary information about the checks (don't output"
        " message information)",
    )

    parser.add_argument(
        '-8', '--no-style-warnings', default=False, action='store_true',
        help="Don't create any warnings about style. This disables the PEP8 tool and"
        " similar checks for formatting."
    )

    tools_help = 'A list of tools to run. Possible values are: %s. By' \
        ' default, the following tools will be run: %s' % (
            ', '.join(sorted(tools.TOOLS.keys())),
            ', '.join(sorted(tools.DEFAULT_TOOLS))
        )
    parser.add_argument(
        '-t', '--tools', default=None, nargs='+', help=tools_help,
    )

    parser.add_argument(
        '-T', '--test-warnings', default=False, action='store_true',
        help="Also check test modules and packages",
    )

    uses_help = 'A list of one or more libraries or frameworks that the' \
        ' project users. Possible values are django, celery. This will be' \
        ' autodetected by default, but if autotectection doesn\'t work,' \
        ' manually specify them using this flag.'
    parser.add_argument(
        '-u', '--uses', help=uses_help, default=[], nargs='+',
    )

    parser.add_argument(
        '-v', '--version', action='store_true',
        help="Print version information and exit",
    )

    parser.add_argument(
        '--absolute-paths', action='store_true', default=False,
        help='Whether to output absolute paths when referencing files in'
        ' messages. By default, paths will be relative to the --path value',
    )

    parser.add_argument(
        '--die-on-tool-error', action='store_true', default=False,
        help='If a tool fails to run, prospector will try to carry on. Use '
        ' this flag to cause prospector to die and raise the exception the '
        ' tool generated. Mostly useful for development on prospector.'
    )

    parser.add_argument(
        '--no-common-plugin', action='store_true', default=False,
    )

    return parser


def _die(message):
    sys.stderr.write('%s\n' % message)
    sys.exit(1)


def run():
    parser = make_arg_parser()
    args = parser.parse_args()

    if args.version:
        sys.stdout.write("Prospector version %s\n" % __pkginfo__.get_version())
        sys.exit(0)

    summary = {
        'started': datetime.now()
    }

    path = args.path or os.path.abspath(os.getcwd())

    try:
        formatter = FORMATTERS[args.output_format]
        summary['formatter'] = args.output_format
    except KeyError:
        _die("Formatter %s is not valid - possible values are %s" % (
            args.output_format,
            ', '.join(FORMATTERS.keys()),
        ))

    libraries_used = []
    profiles = []
    adaptors = []

    if not args.no_common_plugin:
        adaptors.append(CommonAdaptor())
    if not args.no_autodetect:
        for libname, adaptor in autodetect_libraries(path):
            libraries_used.append(libname)
            adaptors.append(adaptor)

    strictness = args.strictness
    strictness_options = ('veryhigh', 'high', 'medium', 'low', 'verylow')
    if strictness not in strictness_options:
        possible = ', '.join(strictness_options)
        _die(
            "%s is not a valid value for strictness - possible values are %s" %
            (strictness, possible)
        )
    else:
        profiles.append('strictness_%s' % strictness)
        summary['strictness'] = strictness

    for library in args.uses:
        if library not in LIBRARY_ADAPTORS:
            possible = ', '.join(LIBRARY_ADAPTORS.keys())
            _die(
                "Library/framework %s is not valid - possible values are %s" %
                (library, possible)
            )
        libraries_used.append(library)
        adaptors.append(LIBRARY_ADAPTORS[library]())

    summary['libraries'] = ', '.join(libraries_used)

    if not args.doc_warnings:
        profiles.append('no_doc_warnings')

    if not args.test_warnings:
        profiles.append('no_test_warnings')

    if args.no_style_warnings:
        profiles.append('no_pep8')

    profiles += args.profiles

    profile_adaptor = ProfileAdaptor(profiles)
    adaptors.append(profile_adaptor)

    summary['adaptors'] = []
    for adaptor in adaptors:
        summary['adaptors'].append(adaptor.name)
    summary['adaptors'] = ', '.join(summary['adaptors'])

    tool_runners = []
    tool_names = args.tools or tools.DEFAULT_TOOLS
    for tool in tool_names:
        if not tool in tools.TOOLS:
            _die("Tool %s is not valid - possible values are %s" % (
                tool,
                ', '.join(tools.TOOLS.keys())
            ))
        if not profile_adaptor.is_tool_enabled(tool):
            continue
        tool_runners.append(tools.TOOLS[tool]())

    summary['tools'] = ', '.join(tool_names)

    ignore = [re.compile(ignore) for ignore in profile_adaptor.profile.ignore]

    for tool in tool_runners:
        tool.prepare(path, ignore, args, adaptors)

    messages = []
    for tool in tool_runners:
        try:
            messages += tool.run()
        except Exception:
            if args.die_on_tool_error:
                raise
            loc = Location(path, None, None, None, None)
            message = "Tool %s failed to run (exception was raised)" % tool.__class__.__name__
            msg = Message(tool.__class__.__name__, 'failure', loc, message=message)
            messages.append(msg)

    for message in messages:
        if args.absolute_paths:
            message.to_absolute_path(path)
        else:
            message.to_relative_path(path)

    if not args.no_blending:
        messages = blender.blend(messages)

    summary['message_count'] = len(messages)
    summary['completed'] = datetime.now()
    delta = (summary['completed'] - summary['started'])
    summary['time_taken'] = '%0.2f' % delta.total_seconds()

    summary['started'] = str(summary['started'])
    summary['completed'] = str(summary['completed'])

    if args.messages_only:
        summary = None
    if args.summary_only:
        messages = None

    formatter(summary, messages)

    sys.exit(0)


if __name__ == '__main__':
    run()



# This module contains the logic for "blending" of errors.
# Since prospector runs multiple tools with overlapping functionality, this module
# exists to merge together equivalent warnings from different tools for the same
# line. For example, both pyflakes and pylint will generate an "Unused Import"
# warning on the same line. This is obviously redundant, so we remove duplicates.
from collections import defaultdict

BLEND = (
    (  # Unused Import
        ('pylint', 'W0611'),
        ('pyflakes', 'FL0001')
    ),
    (  # Syntax Error
        # prefer this error as it will highlight diffs, but not be present for syntax errors not caused by diffs
        ('dodgy', 'diff'),
        ('pylint', 'E0001'),
        ('pyflakes', 'FL9998'),
        ('pep8', 'E901'),
        ('mccabe', 'MC0000')
    ),
    (  # Undefined local variable
        ('pylint', 'E0602'),
        ('pyflakes', 'FL0006')
    ),
    (  # Unused variable
        ('pylint', 'W0612'),
        ('pyflakes', 'FL0013'),
    ),
    (  # Mixed tabs and spaces
        ('pep8', 'E101'),
        ('pylint', 'W0313')
    ),
    (  # Import from __future__ not first import
        ('pylint', 'W0410'),
        ('pyflakes', 'FL0012')
    ),
    (  # Line too long
        ('pep8', 'E501'),
        ('pylint', 'C0301')
    ),
    (  # Trailing whitespace
        ('pep8', 'W291'),
        ('pylint', 'C0303'),
    ),
    (  # No newline at end of file
        ('pep8', 'W292'),
        ('pylint', 'C0304')
    ),
    (  # line ends with semi-colon
        ('pep8', 'E703'),
        ('pylint', 'W0301')
    ),
    (  # multiple statements on one line
        ('pep8', 'E702'),
        ('pylint', 'C0321')
    ),
    (  # incorrect indendation
        ('pep8', 'E111'),
        ('pylint', 'W0311')
    ),
    (  # comma not followed by a space
        ('pep8', 'E231'),
        ('pylint', 'C0324')
    ),
    (  # missing whitespace around operator
        ('pep8', 'E225'),
        ('pylint', 'C0322')
    ),
    (  # missing whitespace around operator
        ('pep8', 'E225'),
        ('pylint', 'C0323')
    ),
    (  # undefined name in __all__
        ('pylint', 'E0603'),
        ('pyflakes', 'FL0008')
    ),
    (  # duplicate argument in function definition
        ('pylint', 'E0108'),
        ('pyflakes', 'FL0010')
    ),
    (  # redefinition of unused function
        ('pyflakes', 'FL0002'),
        ('pylint', 'E0102'),
    )
)

"""
BLEND is a list of tuples of codes to merge together. The earlier codes will
take priority, so if all are found, only the first one will be left after blending.

Note that since not all tools will necessarily be run, the first message for a line
as sorted by the code list will be the one remaining after blending.
"""


def blend_line(messages, blend_combos=BLEND):
    """
    Given a list of messages on the same line, blend them together so that we end
    up with one message per actual problem. Note that we can still return more than
    one message here if there are two or more different errors for the line.
    """
    blend_lists = [[] for _ in range(len(blend_combos))]
    blended = []

    # first we split messages into each of the possible blendable categories
    # so that we have a list of lists of messages which can be blended together
    for message in messages:
        key = (message.source, message.code)
        found = False
        for blend_combo_idx, blend_combo in enumerate(blend_combos):
            if key in blend_combo:
                found = True
                blend_lists[blend_combo_idx].append(message)

        # note: we use 'found=True' here rather than a simple break/for-else because
        # this allows the same message to be put into more than one 'bucket'. This means
        # that the same message from pep8 can 'subsume' two from pylint, for example.

        if not found:
            # if we get here, then this is not a message which can be blended,
            # so by definition is already blended
            blended.append(message)

    # we should now have a list of messages which all represent the same
    # problem on the same line, so we will sort them according to the priority
    # in BLEND and pick the first one
    for blend_combo_idx, blend_list in enumerate(blend_lists):
        if len(blend_list) == 0:
            continue
        blend_list.sort(key=lambda msg: blend_combos[blend_combo_idx].index((msg.source, msg.code)))
        blended.append(blend_list[0])

    return blended


def blend(messages, blend_combos=BLEND):
    # group messages by file and then line number
    msgs_grouped = defaultdict(lambda: defaultdict(list))

    for message in messages:
        msgs_grouped[message.location.path][message.location.line].append(message)

    # now blend together all messages on the same line
    out = []
    for by_line in msgs_grouped.values():
        for messages_on_line in by_line.values():
            out += blend_line(messages_on_line, blend_combos)

    return out
